
#include"Error.h"
#include"Vect.h"
#include"Tensor.h"
#include"NAG.h"

Tensor::Tensor(MatrixKind MA)

{ if (MA==eyes) 
    {
    x=vect3dc(1,0,0);
    y=vect3dc(0,1,0);
    z=vect3dc(0,0,1);
    }
  if (MA==zeros)
    {
    x=vect3dc(0,0,0);
    y=vect3dc(0,0,0);
    z=vect3dc(0,0,0);
    }
 if (MA==ones)
    {
    x=vect3dc(1,1,1);
    y=vect3dc(1,1,1);
    z=vect3dc(1,1,1);
    }
 if (MA==rands)
    {
    x=vect3dc(Complex(drand48(),drand48()),
              Complex(drand48(),drand48()),
              Complex(drand48(),drand48()));
    y=vect3dc(Complex(drand48(),drand48()),
              Complex(drand48(),drand48()),
              Complex(drand48(),drand48()));
    z=vect3dc(Complex(drand48(),drand48()),
              Complex(drand48(),drand48()),
              Complex(drand48(),drand48()));
    }
}

Tensor::Tensor(const Tensor& T)

{ x=T.x;
  y=T.y;
  z=T.z;
}

Tensor::Tensor(vect3dc linex,vect3dc liney,vect3dc linez)

{ x=linex;
  y=liney;
  z=linez;
}

Tensor operator*(const Tensor& T, Complex C)

{ return Tensor(T.x*C, T.y*C, T.z*C);
}

vect3dc operator*(const Tensor& T,vect3dc V)

{ return vect3dc(T.x*V, T.y*V, T.z*V);
}

vect3dc operator*(const Tensor& T,vect3d V)

{ return vect3dc(T.x*V, T.y*V, T.z*V);
}

Tensor operator-(const Tensor& T)

{ return Tensor(-T.x,-T.y,-T.z);
}

Tensor& Tensor::operator+=(const Tensor& T)

{ x=x+T.x;
  y=y+T.y;
  z=z+T.z;
  return (*this);
}

Tensor& Tensor::operator-=(const Tensor& T)

{ x=x-T.x;
  y=y-T.y;
  z=z-T.z;
  return (*this);
}

ostream& operator<<(ostream& O,const Tensor& T)

{ return O<<""<<T.x<<"\n"<<T.y<<"\n"<<T.z<<"";
}

istream& operator>>(istream& O,Tensor& T)
 
{ return O>>T.x>>T.y>>T.z;
}


Tensor prod(vect3dc V1,vect3dc V2)

{return Tensor(V1.x*V2, V1.y*V2, V1.z*V2);
}

Tensor prod(vect3d V1,vect3d V2)

{return Tensor(vect3dc(V1.x*V2,vect3d(0,0,0)), 
               vect3dc(V1.y*V2,vect3d(0,0,0)),
               vect3dc(V1.z*V2,vect3d(0,0,0)));
}

Tensor operator*(const Tensor& T1,const Tensor& T2)

{ Tensor T2t=Tensor(!T2);

  return !Tensor(T1*T2t.x,T1*T2t.y,T1*T2t.z);
}

Tensor operator!(const Tensor& T)

{ return Tensor(vect3dc(T.x.x,T.y.x,T.z.x),
                vect3dc(T.x.y,T.y.y,T.z.y),
                vect3dc(T.x.z,T.y.z,T.z.z));
}

Tensor conj(const Tensor& T)

{ return Tensor(conj(T.x),conj(T.y),conj(T.z));
}

/*****************************************************************************/

SphericalBasis::SphericalBasis()

{ eR=    vect3d(1,0,0);
  eTheta=vect3d(0,1,0);
  ePhi=  vect3d(0,0,1);
}

SphericalBasis::SphericalBasis(double tetha,double phi)

{ eR=    vect3d(sin(tetha)*cos(phi), sin(tetha)*sin(phi), cos(tetha));//eR
  eTheta=vect3d(cos(tetha)*cos(phi), cos(tetha)*sin(phi),-sin(tetha));//eTheta
  ePhi=  vect3d(-sin(phi),           cos(phi),            0);         //ePhi
}

/*****************************************************************************/

Tensor TensorSphericalBasis(Complex tetha,Complex phi)

{ Tensor T;

  T.x=vect3dc(sin(tetha)*cos(phi), sin(tetha)*sin(phi), cos(tetha));  //eR
  T.y=vect3dc(cos(tetha)*cos(phi), cos(tetha)*sin(phi),-sin(tetha));  //eTheta
  T.z=vect3dc(-sin(phi),           cos(phi),            0);           //ePhi
 
  return T;
}

/*****************************************************************************/

CylindricalBasis::CylindricalBasis()

{ eRho=vect3d(1,0,0);
  ePhi=vect3d(0,1,0);
  eZ  =vect3d(0,0,1);
}

CylindricalBasis::CylindricalBasis(double phi)

{ eRho=vect3d( cos(phi), sin(phi), 0);
  ePhi=vect3d(-sin(phi), cos(phi), 0);
  eZ  =vect3d( 0       , 0       , 1);
}

/*****************************************************************************/

CoordinateSystem::CoordinateSystem()

{ e1=vect3d(1,0,0);
  e2=vect3d(0,1,0);
  e3=vect3d(0,0,1);
}

CoordinateSystem::CoordinateSystem(const CoordinateSystem& cs)

{ e1=cs.e1;
  e2=cs.e2;
  e3=cs.e3;
}

CoordinateSystem::CoordinateSystem(const CylindricalBasis& cb)

{ e1=cb.eRho;
  e2=cb.ePhi;
  e3=cb.eZ;
}

CoordinateSystem::CoordinateSystem(const SphericalBasis& sb)

{ e1=sb.eR;
  e2=sb.eTheta;
  e3=sb.ePhi;
}

CoordinateSystem::CoordinateSystem(double alpha,double beta,double gamma)

{ e1=vect3d(   cos(gamma)*cos(beta)*cos(alpha)-sin(gamma)*sin(alpha),
              -sin(gamma)*cos(beta)*cos(alpha)-cos(gamma)*sin(alpha),
                          sin(beta)*cos(alpha));

  e2=vect3d(   cos(gamma)*cos(beta)*sin(alpha)+sin(gamma)*cos(alpha),
              -sin(gamma)*cos(beta)*sin(alpha)+cos(gamma)*cos(alpha),
                          sin(beta)*sin(alpha));

  e3=vect3d(  -cos(gamma)*sin(beta),
               sin(gamma)*sin(beta),
                          cos(beta));
}

vect3d CoordinateSystem::ITransform(const vect3d& V)

{// return vect3d(e1.x*V.x + e2.x*V.y + e3.x*V.z,      // [e1;e2;e3]*V
 //               e1.y*V.x + e2.y*V.y + e3.y*V.z,
 //               e1.z*V.x + e2.z*V.y + e3.z*V.z);
  return real(M()*V);
}

vect3dc CoordinateSystem::ITransform(const vect3dc& V)

{// return vect3dc(e1.x*V.x + e2.x*V.y + e3.x*V.z,     // [e1;e2;e3]*V
 //                e1.y*V.x + e2.y*V.y + e3.y*V.z,
 //                e1.z*V.x + e2.z*V.y + e3.z*V.z);
  return M()*V;
}

vect3d CoordinateSystem::Transform(const vect3d& V)

{// return vect3d(e1*V,e2*V,e3*V);                     // [e1;e2;e3]t*V
  return real((!M())*V);
}

vect3dc CoordinateSystem::Transform(const vect3dc& V)

{// return vect3dc(e1*V,e2*V,e3*V);                    // [e1;e2;e3]t*V
  return (!M())*V;
}

Tensor CoordinateSystem::M()

{ vect3d Zeros=vect3d(0,0,0);

  return !Tensor(vect3dc(e1,Zeros),
                 vect3dc(e2,Zeros),
                 vect3dc(e3,Zeros));
}

Tensor CoordinateSystem::Transform(const Tensor& T)

{ return (!M())*T*M();
}

Tensor CoordinateSystem::ITransform(const Tensor& T)

{ return M()*T*(!M());
}

/*****************************************************************************/

vect3dc Tensor::ITransform(const vect3dc& V)

{ return (Inv(*this))*V;
}

vect3dc Tensor::Transform(const vect3dc& V)

{ return (*this)*V;
}

Tensor Tensor::Transform(const Tensor& T)

{ return (*this)*T*(Inv(*this));
}

Tensor Tensor::ITransform(const Tensor& T)

{ return (Inv(*this))*T*(*this);
}

/*****************************************************************************/

TENSOR::TENSOR(const TENSOR& t) 

{ Scalar=t.Scalar;
  T=t.T;
  C=t.C;
}

TENSOR::TENSOR(const Tensor& t) 

{ Scalar=0;
  T=t;
}

TENSOR::TENSOR(const Complex& c)

{ Scalar=1;
  C=c;
}

Tensor TENSOR::GetTensor() const

{ if (Scalar) return C*Tensor(eyes);
  return T;
}

Complex TENSOR::GetScalar() const    

{ if (!Scalar) FatalError("attempt to get a scalar from a tensorial tensor");
  return C;
}

int TENSOR::TryToMakeScalar()

{ if(Scalar) return 1;
  double diff=Abs(vect3d(Abs(T.x),Abs(T.y),Abs(T.z)))/1E15;;

  if ((Abs(T.x.y)>diff)||(Abs(T.x.z)>diff)||
      (Abs(T.y.x)>diff)||(Abs(T.y.z)>diff)||
      (Abs(T.z.x)>diff)||(Abs(T.z.y)>diff)||
      (Abs(T.x.x-T.y.y)>diff) ||
      (Abs(T.x.x-T.z.z)>diff)) return 0;
			
  (*this)=TENSOR(T.x.x);
  return 1;
}

TENSOR operator*(const TENSOR& T1, Complex C)

{ if(T1.Scalar) return TENSOR(T1.GetScalar()*C);
  return  TENSOR(T1.GetTensor()*C);
}


vect3dc operator*(const TENSOR& T1,vect3d V)

{ if(T1.Scalar) return T1.GetScalar()*V;
  return  T1.GetTensor()*V;
}


vect3dc operator*(const TENSOR& T1,vect3dc V)

{ if(T1.Scalar) return T1.GetScalar()*V;
  return  T1.GetTensor()*V;
}

TENSOR  operator*(const TENSOR& T1,const TENSOR& T2)

{ if (( T1.Scalar)&&( T2.Scalar)) return TENSOR(T1.GetScalar()*T2.GetScalar());
  if (( T1.Scalar)&&(!T2.Scalar)) return TENSOR(T1.GetScalar()*T2.GetTensor());
  if ((!T1.Scalar)&&( T2.Scalar)) return TENSOR(T1.GetTensor()*T2.GetScalar());

  return T1.GetTensor()*T2.GetTensor();
}

TENSOR  operator-(const TENSOR& T1)

{ if ( T1.Scalar) return TENSOR(-T1.GetScalar());
  return TENSOR(-T1.GetTensor());
}

TENSOR& TENSOR::operator+=(const TENSOR& T1)

{ if ((Scalar)&&(T1.Scalar)) C+=T1.C;
  else (*this)=TENSOR(GetTensor()+T1.GetTensor());
  return (*this);
}

TENSOR& TENSOR::operator-=(const TENSOR& T1)

{ if ((Scalar)&&(T1.Scalar)) C-=T1.C;
  else (*this)=TENSOR(GetTensor()-T1.GetTensor());
  return (*this);
}

TENSOR operator!(const TENSOR& T1)

{ if (T1.Scalar) return T1;
  return TENSOR(!T1.GetTensor());
}

TENSOR conj(const TENSOR& T)

{ if (T.Scalar) return conj(T.C);
  return TENSOR(conj(T.GetTensor()));
}

int Inv(TENSOR& Result,TENSOR T1)             //return 0: no error

{ int err=0;
  if(T1.Scalar)
    {
    Complex c=(T1.GetScalar());
    if (c==0) err=1;
    Result=TENSOR(1.0/c);
    }
  else
    { 
    Tensor t;
    err=Inv(t,T1.GetTensor());
    Result=TENSOR(t);
    }
  return err;
};

/*****************************************************************************/

int Inv(Tensor& Result,Tensor T)

{ int     N=3;
  int     IPIV[3];
  int     Info=0;
  Complex Work[6];
  int     LWork=6;

  FortranCall(f07arf)(N,N,&T.x.x,N,IPIV,Info);

  if(Info!=0)
    {
    if (Info<0) FatalError("Illegal paramter for procedure call f07arf");
    else return 1;
    }
  FortranCall(f07awf)(N,&T.x.x,N,IPIV,Work,LWork,Info);
  if(Info!=0)
    {
    if (Info<0) FatalError("Illegal paramter for procedure call f07awf");
    else return 1;
    }
  Result=T;
  return 0;
}

Tensor Inv(Tensor T)

{ Tensor Result;

  int err=Inv(Result,T);
  if (err) cerr<<"Error during in invertion of Tensor\n";
  return Result;
}


/*****************************************************************************/
