#if !defined _TENSOR_
#define _TENSOR_

#include<Complex.h>
#include"Object.h"
#include"Error.h"
#include"Math.h"
#include"Vect.h"

typedef enum MatrixKind{ones,zeros,eyes,rands};

class Tensor{
 public:
  vect3dc x;   // first  line
  vect3dc y;   // second line
  vect3dc z;   // third  line
 public:
  Tensor(const Tensor& T);
  Tensor(MatrixKind MA);
  Tensor(vect3dc linex,vect3dc liney,vect3dc linez);
  Tensor(){(*this)=Tensor(zeros);};
  ~Tensor(){};
  friend Tensor  operator*(const Tensor& T, Complex C);
  friend Tensor  operator*(Complex C, const Tensor& T){ return T*C;};
  friend vect3dc operator*(const Tensor& T,vect3d V);
  friend vect3dc operator*(const Tensor& T,vect3dc V);
  friend Tensor  operator*(const Tensor& T1,const Tensor& T2);
  friend Tensor  operator-(const Tensor& T);
  Tensor& operator+=(const Tensor& T);
  Tensor& operator-=(const Tensor& T);
  friend Tensor operator!(const Tensor& T);
  friend Tensor operator+(Tensor T1,const Tensor& T2){ return T1+=T2;}; 
  friend Tensor operator-(Tensor T1,const Tensor& T2){ return T1-=T2;}; 
  friend ostream& operator<<(ostream& O,const Tensor& T);
  friend istream& operator>>(istream& O,Tensor& T);
  friend Tensor prod(vect3dc V1,vect3dc V2);
  friend Tensor prod(vect3d  V1,vect3d  V2);
  friend int Inv(Tensor& Result,Tensor T);             //return 0: no error
  friend Tensor Inv(Tensor T);
  friend Tensor conj(const Tensor& T);

  vect3dc Transform(const vect3dc& V);           //return Mt*V
  vect3dc ITransform(const vect3dc& V);          //return M*V
  Tensor  Transform(const Tensor& T);            //return Mt*T*M
  Tensor  ITransform(const Tensor& T);           //return M*T*Mt
};  
  
/*****************************************************************************/
class TENSOR : public Object{
 protected:
  int Scalar;
  Tensor T;
  Complex C;
 public:
  TENSOR(){(*this)=TENSOR(Complex(0,0));};
  TENSOR(const TENSOR& t) ;
  TENSOR(const Tensor& t) ;
  TENSOR(const Complex& c);
  int     IsScalar()   const {return Scalar;};
  int     IsTensorial()const {return !Scalar;};
  Tensor  GetTensor()  const;
  Complex GetScalar()  const;
  int TryToMakeScalar();

  friend TENSOR  operator*(const TENSOR& T, Complex C);
  friend TENSOR  operator*(Complex C, const TENSOR& T){ return T*C;};
  friend vect3dc operator*(const TENSOR& T,vect3d V);
  friend vect3dc operator*(const TENSOR& T,vect3dc V);
  friend TENSOR  operator*(const TENSOR& T1,const TENSOR& T2);
  friend TENSOR  operator-(const TENSOR& T);
  TENSOR& operator+=(const TENSOR& T);
  TENSOR& operator-=(const TENSOR& T);
  friend TENSOR operator!(const TENSOR& T);
  friend TENSOR operator+(TENSOR T1,const TENSOR& T2){ return T1+=T2;}; 
  friend TENSOR operator-(TENSOR T1,const TENSOR& T2){ return T1-=T2;}; 
  friend int Inv(TENSOR& Result,TENSOR T);             //return 0: no error
  friend TENSOR conj(const TENSOR& T);

};
 
/*****************************************************************************/

class SphericalBasis{
public:
  SphericalBasis();
  SphericalBasis(double tetha,double phi);
  vect3d eR, eTheta, ePhi;
};

class CylindricalBasis{
public:
  CylindricalBasis();
  CylindricalBasis(double phi);
  vect3d eRho, ePhi, eZ;
};

/*****************************************************************************/

class CoordinateSystem{
public:
  vect3d e1,e2,e3;    //New basis vectors exprimated in the old basis
  Tensor  M();
public:
  CoordinateSystem();
  CoordinateSystem(vect3d E1,vect3d E2,vect3d E3){e1=E1;e2=E2;e3=E3;};
  CoordinateSystem(vect3d E1,vect3d E2){e1=E1;e2=E2;e3=e1^e2;};
  CoordinateSystem(const CoordinateSystem& cs);
  CoordinateSystem(const CylindricalBasis& cb);  //e1=eRho; e2=ePhi;  e3=cb.eZ;
  CoordinateSystem(const SphericalBasis& sb);    //e1=eR;   e2=eTheta;e3=ePhi;
  CoordinateSystem(double alpha,double beta,double gamma);
                                                 //Arfken p.200
  vect3d  Transform(const vect3d& V);            //return Mt*V
  vect3dc Transform(const vect3dc& V);
  vect3d  ITransform(const vect3d& V);           //return M*V
  vect3dc ITransform(const vect3dc& V);
  Tensor  Transform(const Tensor& T);            //return Mt*T*M
  Tensor  ITransform(const Tensor& T);           //return M*T*Mt
  void Norm(){e1=e1/Abs(e1);e2=e2/Abs(e2);e3=e3/Abs(e3);};
};

Tensor TensorSphericalBasis(Complex tetha,Complex phi);

//  T.x:eR
//  T.y:eTheta
//  T.z:ePhi

/*****************************************************************************/

#endif


