/***************************************************************************/
/*                                                                         */
/*   Untersuchung eines neuen Verfahrens zur Berechnung von Streufeldern.  */
/*                   The Mesaured Equation of Invariance                   */
/*                                                                         */
/***************************************************************************/
/*   File : Vect.c                                                         */
/*                                                                         */
/*   Geschrieben : 04.07.95 Nicolas Piller                                 */
/*   Aenderungen : ...      ...                                            */
/*                                                                         */
/***************************************************************************/

#include <Complex.h>
#include <iostream.h>
#include "Const.h"
#include "Vect.h"

//*********************************2D********************************* 

VectorOutputMethodT VectorOutputMethod=MatlabVectorOutputMethod;
VectorInputMethodT  VectorInputMethod= MatlabVectorInputMethod;

#define OP222add(T1,T2,T3)   					\
T1 operator+(T2 V1, T3 V2)					\
{ return T1(V1.x + V2.x, V1.y + V2.y);				\
}

#define OP222sub(T1,T2,T3)   					\
T1 operator-(T2 V1, T3 V2)					\
{ return T1(V1.x - V2.x, V1.y - V2.y);				\
}

#define OP22neg(T1,T2) 		    				\
T1 operator-(T2 V1)						\
{ return T1(- V1.x, -V1.y);					\
}

#define OP122dot(T1,T2,T3) 		    			\
T1 operator*(T2 V1, T3 V2)					\
{ return T1(V1.x * V2.x + V1.y * V2.y);				\
}

#define OP212mult(T1,T2,T3) 	    				\
T1 operator*(T2 V1, T3 V2)					\
{ return T1(V1 * V2.x,V1 * V2.y);				\
}

#define OP221mult(T1,T2,T3) 	    				\
T1 operator*(T2 V1, T3 V2)					\
{ return T1(V1.x * V2,V1.y * V2);				\
}

#define OP221div(T1,T2,T3) 	    				\
T1 operator/(T2 V1, T3 V2)					\
{ return T1(V1.x / V2,V1.y / V2);				\
}

#define NORM12(T1,T2)						\
T1 norm(T2 V1)							\
{return sqrt(real(V1.x)*real(V1.x) + imag(V1.x)*imag(V1.x) +	\
             real(V1.y)*real(V1.y) + imag(V1.y)*imag(V1.y));	\
}

#define UNIT22(T1,T2)						\
T1 unit(T2 V1)							\
{return V1/norm(V1);						\
}

#define FCT22(FCT,T1,T2)        T1				\
FCT(T2 V1)							\
{return T1(FCT(V1.x),FCT(V1.y));				\
}

#define OPPMEQ22(OP,T1,T2,T3)	T1 				\
T2 :: operator OP (T3 V1)					\
{ x OP V1.x;							\
  y OP V1.y;							\
  return *this;							\
}								\

#define OPEQ2(T1)						\
int operator==(T1 V1,T1 V2)					\
{return ((V1.x==V2.x)&&(V1.y==V2.y));}				

#define OPNEQ2(T1)						\
int operator!=(T1 V1,T1 V2)					\
{return !((V1.x==V2.x)&&(V1.y==V2.y));}				


ostream& operator<<(ostream& o,const vect2d& A)	

{ if (VectorOutputMethod==MMPVectorOutputMethod)                
  return o<< "  " << A.x << " " << A.y << " " ; 			

  return o<< "[" << A.x << ";" << A.y << "]" ; 			
}

ostream& operator<<(ostream& o,const vect3d& A)	

{ if (VectorOutputMethod==MMPVectorOutputMethod)                
  return o<< "  " << A.x << " " << A.y << " " << A.z << " " ; 	

  return o<< "[" << A.x << ";" << A.y << ";" << A.z << "]" ;
}

ostream& operator<<(ostream& o,const vect2dc& A)	

{ if (VectorOutputMethod==MMPVectorOutputMethod)                
  return o<< "  " << real(A.x) << " " << imag(A.x) << " "
          << " "<< real(A.y) << " " << imag(A.y) << " ";

  return o<< "[" << A.x << ";" << A.y << "]" ; 			
}

ostream& operator<<(ostream& o,const vect3dc& A)	

{ if (VectorOutputMethod==MMPVectorOutputMethod)                
  return o<< "  " << real(A.x) << " " << imag(A.x) << " "
          << " "<< real(A.y) << " " << imag(A.y) << " "
          << " "<< real(A.z) << " " << imag(A.z) << " ";

  return o<< "[" << A.x << ";" << A.y << ";" << A.z << "]" ;
}


#define ISTREAM2(T1)  char c;					\
  {in >>c;} while ((c==' ')||(c=='\n'));			\
  if(c!='[') in.clear(ios::badbit);				\
  in >>L.x;							\
  {in >>c;} while ((c==' ')||(c=='\n'));			\
  if(c!=';') in.clear(ios::badbit);				\
  in >>L.y;							\
  {in >>c;} while ((c==' ')||(c=='\n'));			\
  if(c!=']') in.clear(ios::badbit);				\
  return in;

istream& operator>>(istream& in,vect2d& L)

{ if (VectorInputMethod==MMPVectorInputMethod)
    {
    in>>L.x>>L.y;
    return in;
    }
  ISTREAM2(vect2d);
}

istream& operator>>(istream& in,vect2dc& L)

{ if (VectorInputMethod==MMPVectorInputMethod)
    {
    double r,i;
    in>>r>>i;
    L.x=Complex(r,i);
    in>>r>>i;
    L.y=Complex(r,i);
    return in;
    }
  ISTREAM2(vect2dc);
}


//*********************************VECT2D********************************* 


//*********************************Constructor, Destructor

vect2d::vect2d()

{ x=0;
  y=0;
}

vect2d::vect2d(double X,double Y)

{ x=X;
  y=Y;
}

vect2d MakeVect2D(vect3d V3D)

{ return vect2d(V3D.x,V3D.y);
}

vect2d::~vect2d()

{}

//*********************************Methodes

double norm (vect2d V1)

{ return sqrt(V1.x*V1.x + V1.y*V1.y);}
 
UNIT22(vect2d,vect2d)
double Abs(vect2d V1){return norm(V1);}


//*********************************Operator

OP222add(vect2d,vect2d,vect2d)
OP222sub(vect2d,vect2d,vect2d)
OP22neg(vect2d,vect2d)
OP122dot(double,vect2d,vect2d)
OP212mult(vect2d,double,vect2d)
OP221mult(vect2d,vect2d,double)
OP221div(vect2d,vect2d,double)

OPPMEQ22(+=,vect2d&,vect2d,vect2d)
OPPMEQ22(-=,vect2d&,vect2d,vect2d)

OPEQ2(vect2d)
OPNEQ2(vect2d)

//*********************************VECT2DC*********************************

//*********************************Constructor, Destructor

vect2dc::vect2dc()

{ x=Complex(0,0);
  y=Complex(0,0);
}

vect2dc::vect2dc(Complex X,Complex Y)

{ x=X;
  y=Y;
}

vect2dc::vect2dc(vect2d Re,vect2d Im)

{ *this =  Re + Complex(0,1)* Im;
}

vect2dc MakeVect2D(vect3dc V3D)

{ return vect2dc(V3D.x,V3D.y);
}

vect2dc::~vect2dc()

{}

//*********************************Methodes

NORM12(double,vect2dc)
UNIT22(vect2dc,vect2dc)
FCT22(conj,vect2dc,vect2dc)
FCT22(real,vect2d,vect2dc)
FCT22(imag,vect2d,vect2dc)
double Abs(vect2dc V){return norm(V);}

//*********************************Operator

OP222add(vect2dc,vect2dc,vect2dc)
OP222add(vect2dc,vect2d,vect2dc)
OP222add(vect2dc,vect2dc,vect2d)
OP222sub(vect2dc,vect2dc,vect2dc)
OP222sub(vect2dc,vect2d,vect2dc)
OP222sub(vect2dc,vect2dc,vect2d)

OP22neg(vect2dc,vect2dc)

OP122dot(Complex,vect2dc,vect2dc)
OP122dot(Complex,vect2dc,vect2d)
OP122dot(Complex,vect2d,vect2dc)
OP212mult(vect2dc,double,vect2dc)
OP212mult(vect2dc,Complex,vect2d)
OP212mult(vect2dc,Complex,vect2dc)

OP221mult(vect2dc,vect2dc,double)
OP221mult(vect2dc,vect2dc,Complex)
OP221mult(vect2dc,vect2d,Complex)
OP221div(vect2dc,vect2dc,double)
OP221div(vect2dc,vect2dc,Complex)
OP221div(vect2dc,vect2d,Complex)

OPPMEQ22(+=,vect2dc&,vect2dc,vect2dc)
OPPMEQ22(-=,vect2dc&,vect2dc,vect2dc)
OPPMEQ22(+=,vect2dc&,vect2dc,vect2d)
OPPMEQ22(-=,vect2dc&,vect2dc,vect2d)

OPEQ2(vect2dc)
OPNEQ2(vect2dc)

//*********************************3D********************************* 

#define OP333add(T1,T2,T3)   					\
T1 operator+(T2 V1, T3 V2)					\
{ return T1(V1.x + V2.x, V1.y + V2.y, V1.z + V2.z);		\
}

#define OP333sub(T1,T2,T3)   					\
T1 operator-(T2 V1, T3 V2)					\
{ return T1(V1.x - V2.x, V1.y - V2.y, V1.z - V2.z);		\
}

#define OP33neg(T1,T2) 		    				\
T1 operator-(T2 V1)						\
{ return T1(- V1.x, -V1.y, -V1.z);				\
}

#define OP133dot(T1,T2,T3) 		    			\
T1 operator*(T2 V1, T3 V2)					\
{ return T1(V1.x * V2.x + V1.y * V2.y + V1.z * V2.z);		\
}

#define OP313mult(T1,T2,T3) 	    				\
T1 operator*(T2 V1, T3 V2)					\
{ return T1(V1 * V2.x, V1 * V2.y, V1 * V2.z);			\
}

#define OP331mult(T1,T2,T3) 	    				\
T1 operator*(T2 V1, T3 V2)					\
{ return T1(V1.x * V2, V1.y * V2, V1.z * V2);			\
}

#define OP331div(T1,T2,T3) 	    				\
T1 operator/(T2 V1, T3 V2)					\
{ return T1(V1.x / V2,V1.y / V2,V1.z / V2);			\
}

#define OP333cross(T1,T2,T3)    T1				\
operator^(T2 V1,T3 V2)						\
{ return T1(V1.y*V2.z - V1.z*V2.y,				\
            V1.z*V2.x - V1.x*V2.z,				\
            V1.x*V2.y - V1.y*V2.x);}

#define NORM13(T1,T2)						\
T1 norm(T2 V1)							\
{return sqrt(real(V1.x)*real(V1.x) + imag(V1.x)*imag(V1.x) +	\
             real(V1.y)*real(V1.y) + imag(V1.y)*imag(V1.y) +	\
             real(V1.z)*real(V1.z) + imag(V1.z)*imag(V1.z));	\
}

#define UNIT33(T1,T2)						\
T1 unit(T2 V1)							\
{return V1/norm(V1);						\
}

#define FCT33(FCT,T1,T2)        T1				\
FCT(T2 V1)							\
{return T1(FCT(V1.x),FCT(V1.y),FCT(V1.z));			\
}

#define OPPMEQ33(OP,T1,T2,T3)	T1 				\
T2 :: operator OP (T3 V1)					\
{ x OP V1.x;							\
  y OP V1.y;							\
  z OP V1.z;							\
  return *this;							\
}								\

#define OPEQ3(T1)						\
int operator==(T1 V1,T1 V2)					\
{return ((V1.x==V2.x)&&(V1.y==V2.y)&&(V1.z==V2.z));}	       		

#define OPNEQ3(T1)						\
int operator!=(T1 V1,T1 V2)					\
{return !((V1.x==V2.x)&&(V1.y==V2.y)&&(V1.z==V2.z));}	       		

#define ISTREAM3(T1)     char c;				\
  {in >>c;} while ((c==' ')||(c=='\n'));			\
  if(c!='[') in.clear(ios::badbit);				\
  in >>L.x;							\
  {in >>c;} while ((c==' ')||(c=='\n'));			\
  if(c!=';') in.clear(ios::badbit);     			\
  in >>L.y;							\
  {in >>c;} while ((c==' ')||(c=='\n'));			\
  if(c!=';') in.clear(ios::badbit);      			\
  in >>L.z;							\
  {in >>c;} while ((c==' ')||(c=='\n'));			\
  if(c!=']') in.clear(ios::badbit);      			\
  return in;							


istream& operator>>(istream& in,vect3d& L)

{ if (VectorInputMethod==MMPVectorInputMethod)
    {
    in>>L.x>>L.y>>L.z;
    return in;
    }
  ISTREAM3(vect3d);
}


istream& operator>>(istream& in,vect3dc& L)

{ if (VectorInputMethod==MMPVectorInputMethod)
    {
    double r,i;
    in>>r>>i;
    L.x=Complex(r,i);
    in>>r>>i;
    L.y=Complex(r,i);
    in>>r>>i;
    L.z=Complex(r,i);
    return in;
    }
  ISTREAM3(vect3dc);
}

//*********************************VECT3D********************************* 


//*********************************Constructor, Destructor

vect3d::vect3d()

{ x=0;
  y=0;
  z=0;
}

vect3d::vect3d(double X,double Y,double Z)

{ x=X;
  y=Y;
  z=Z;
}

vect3d::vect3d(vect2d V2D)

{ x=V2D.x;
  y=V2D.y;
  z=0;
}

vect3d::~vect3d()

{}

//*********************************Methodes

double norm(vect3d V1)

{ return sqrt(V1.x*V1.x + V1.y*V1.y + V1.z*V1.z);}

double Abs(vect3d V1){return norm(V1);}
  
UNIT33(vect3d,vect3d)
//*********************************Operator

OP333add(vect3d,vect3d,vect3d)
OP333sub(vect3d,vect3d,vect3d)
OP33neg(vect3d,vect3d)
OP133dot(double,vect3d,vect3d)
OP333cross(vect3d,vect3d,vect3d)
OP313mult(vect3d,double,vect3d)
OP331mult(vect3d,vect3d,double)
OP331div(vect3d,vect3d,double)

OPPMEQ33(+=,vect3d&,vect3d,vect3d)
OPPMEQ33(-=,vect3d&,vect3d,vect3d)

OPEQ3(vect3d)
OPNEQ3(vect3d)

//*********************************VECT3DC*********************************

//*********************************Constructor, Destructor

vect3dc::vect3dc()

{ x=Complex(0,0);
  y=Complex(0,0);
  z=Complex(0,0);
}

vect3dc::vect3dc(Complex X,Complex Y,Complex Z)

{ x=X;
  y=Y;
  z=Z;
}

vect3dc::vect3dc(vect3d Re,vect3d Im)

{ *this =  Re + Complex(0,1)* Im;
}

vect3dc::vect3dc(vect2dc V3D)

{ x=V3D.x;
  y=V3D.y;
  z=Complex(0,0);
}

vect3dc::~vect3dc()

{}

//*********************************Methodes

NORM13(double,vect3dc)
UNIT33(vect3dc,vect3dc)
FCT33(real,vect3d,vect3dc)
FCT33(imag,vect3d,vect3dc)
FCT33(conj,vect3dc,vect3dc)
double Abs(vect3dc v){return norm(v);}

//*********************************Operator

OP333add(vect3dc,vect3dc,vect3dc)
OP333add(vect3dc,vect3d,vect3dc)
OP333add(vect3dc,vect3dc,vect3d)
OP333sub(vect3dc,vect3dc,vect3dc)
OP333sub(vect3dc,vect3d,vect3dc)
OP333sub(vect3dc,vect3dc,vect3d)

OP33neg(vect3dc,vect3dc)

OP133dot(Complex,vect3dc,vect3dc)
OP133dot(Complex,vect3dc,vect3d)
OP133dot(Complex,vect3d,vect3dc)
OP333cross(vect3dc,vect3dc,vect3dc)
OP333cross(vect3dc,vect3dc,vect3d)
OP333cross(vect3dc,vect3d,vect3dc)
OP313mult(vect3dc,double,vect3dc)
OP313mult(vect3dc,Complex,vect3d)
OP313mult(vect3dc,Complex,vect3dc)

OP331mult(vect3dc,vect3dc,double)
OP331mult(vect3dc,vect3dc,Complex)
OP331mult(vect3dc,vect3d,Complex)
OP331div(vect3dc,vect3dc,double)
OP331div(vect3dc,vect3dc,Complex)
OP331div(vect3dc,vect3d,Complex)

OPPMEQ33(+=,vect3dc&,vect3dc,vect3dc)
OPPMEQ33(-=,vect3dc&,vect3dc,vect3dc)
OPPMEQ33(+=,vect3dc&,vect3dc,vect3d)
OPPMEQ33(-=,vect3dc&,vect3dc,vect3d)

OPEQ3(vect3dc)
OPNEQ3(vect3dc)

//*********************************Operator

ostream& mmpwrite(ostream& o,const Complex& C) 

{ return o<<real(C)<<" "<<imag(C)<<" ";
};

ostream& mmpwrite(ostream& o,const vect2d& C)

{ return o<<C.x<<" "<<C.y<<" ";
};

ostream& mmpwrite(ostream& o,const vect3d& C)
    
{ return o<<C.x<<" "<<C.y<<" "<<C.z<<" ";
};
 
ostream& mmpwrite(ostream& o,const vect2dc& C)

{ mmpwrite(o,C.x)<<" ";
  mmpwrite(o,C.y)<<" ";
  return o;
};

ostream& mmpwrite(ostream& o,const vect3dc& C)
    
{ mmpwrite(o,C.x)<<" ";
  mmpwrite(o,C.y)<<" ";
  mmpwrite(o,C.z)<<" ";
  return o;
};

//*********************************Operator

int vect3dc::IsNormalToXYPlane(double accrel)

{ vect3d nz=vect3d(0,0,1);
  vect3dc dv= (*this)-((*this)*nz)*nz;

  if (Abs(dv)<accrel*Abs((*this))) return 1;
  return 0;
}

int vect3dc::IsNormalToXYPlane()

{ return this->IsNormalToXYPlane(0.0001);
}

void vect3dc::MakeNormalToXYPlane()

{ vect3d nz=vect3d(0,0,1);

  (*this)=((*this)*nz)*nz;
}

int vect3dc::IsParallelToXYPlane(double accrel)

{ vect3d nz=vect3d(0,0,1);
  vect3dc dv=((*this)*nz)*nz;

  if (Abs(dv)<accrel*Abs((*this))) return 1;
  return 0;
}

int vect3dc::IsParallelToXYPlane()

{ return this->IsParallelToXYPlane(0.0001);
}

void vect3dc::MakeParallelToXYPlane()

{ vect3d nz=vect3d(0,0,1);
  vect3dc dv=((*this)*nz)*nz;

  (*this)=(*this)-dv;
}

